@@ -24,26 +24,15 @@ THE SOFTWARE.
 #ifndef HIP_INCLUDE_HIP_AMD_DETAIL_HIP_FP16_H
 #define HIP_INCLUDE_HIP_AMD_DETAIL_HIP_FP16_H
 
-#if defined(__clang__)
-#pragma clang diagnostic push
-#pragma clang diagnostic ignored "-Wreserved-identifier"
-#pragma clang diagnostic ignored "-Wreserved-macro-identifier"
-#pragma clang diagnostic ignored "-Wc++98-compat"
-#pragma clang diagnostic ignored "-Wc++98-compat-pedantic"
-#pragma clang diagnostic ignored "-Wsign-conversion"
-#pragma clang diagnostic ignored "-Wfloat-conversion"
-#pragma clang diagnostic ignored "-Wdouble-promotion"
-#pragma clang diagnostic ignored "-Wnested-anon-types"
-#pragma clang diagnostic ignored "-Wgnu-anonymous-struct"
-#pragma clang diagnostic ignored "-Wfloat-equal"
-#endif
-
 #if defined(__HIPCC_RTC__)
   #define __HOST_DEVICE__ __device__
 #else
   #define __HOST_DEVICE__ __host__ __device__
   #include <hip/amd_detail/amd_hip_common.h>
   #include "hip/amd_detail/host_defines.h"
+#if defined(__clang__) && defined(__HIP__)
+  #include "hip/amd_detail/amd_hip_atomic.h"
+#endif // defined(__clang__) && defined(__HIP__)
   #include <assert.h>
   #if defined(__cplusplus)
     #include <algorithm>
@@ -68,11 +57,11 @@ THE SOFTWARE.
         union {
             static_assert(sizeof(_Float16_2) == sizeof(unsigned short[2]), "");
 
-            _Float16_2 data;
             struct {
-                unsigned short x;
-                unsigned short y;
+                __half_raw x;
+                __half_raw y;
             };
+            _Float16_2 data;
         };
     };
 
@@ -355,26 +344,25 @@ THE SOFTWARE.
                 static_assert(
                     sizeof(_Float16_2) == sizeof(unsigned short[2]), "");
 
-                _Float16_2 data;
                 struct {
-                    unsigned short x;
-                    unsigned short y;
+                    __half x;
+                    __half y;
                 };
+                _Float16_2 data;
             };
 
             // CREATORS
             __HOST_DEVICE__
             __half2() = default;
             __HOST_DEVICE__
-            __half2(const __half2_raw& x) : data{x.data} {}
+            __half2(const __half2_raw& xx) : data{xx.data} {}
             __HOST_DEVICE__
-            __half2(decltype(data) x) : data{x} {}
+            __half2(decltype(data) xx) : data{xx} {}
             __HOST_DEVICE__
-            __half2(const __half& x, const __half& y)
+            __half2(const __half& xx, const __half& yy)
                 :
-                data{
-                    static_cast<__half_raw>(x).data,
-                    static_cast<__half_raw>(y).data}
+                data{static_cast<__half_raw>(xx).data,
+                     static_cast<__half_raw>(yy).data}
             {}
             __HOST_DEVICE__
             __half2(const __half2&) = default;
@@ -389,36 +377,36 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2& operator=(__half2&&) = default;
             __HOST_DEVICE__
-            __half2& operator=(const __half2_raw& x)
+            __half2& operator=(const __half2_raw& xx)
             {
-                data = x.data;
+                data = xx.data;
                 return *this;
             }
 
             // MANIPULATORS - DEVICE ONLY
             #if !defined(__HIP_NO_HALF_OPERATORS__)
                 __device__
-                __half2& operator+=(const __half2& x)
+                __half2& operator+=(const __half2& xx)
                 {
-                    data += x.data;
+                    data += xx.data;
                     return *this;
                 }
                 __device__
-                __half2& operator-=(const __half2& x)
+                __half2& operator-=(const __half2& xx)
                 {
-                    data -= x.data;
+                    data -= xx.data;
                     return *this;
                 }
                 __device__
-                __half2& operator*=(const __half2& x)
+                __half2& operator*=(const __half2& xx)
                 {
-                    data *= x.data;
+                    data *= xx.data;
                     return *this;
                 }
                 __device__
-                __half2& operator/=(const __half2& x)
+                __half2& operator/=(const __half2& xx)
                 {
-                    data /= x.data;
+                    data /= xx.data;
                     return *this;
                 }
                 __device__
@@ -445,7 +433,11 @@ THE SOFTWARE.
             __HOST_DEVICE__
             operator decltype(data)() const { return data; }
             __HOST_DEVICE__
-            operator __half2_raw() const { return __half2_raw{data}; }
+            operator __half2_raw() const {
+              __half2_raw r;
+              r.data = data;
+              return r;
+            }
 
             // ACCESSORS - DEVICE ONLY
             #if !defined(__HIP_NO_HALF_OPERATORS__)
@@ -465,74 +457,74 @@ THE SOFTWARE.
                 friend
                 inline
                 __device__
-                __half2 operator+(const __half2& x, const __half2& y)
+                __half2 operator+(const __half2& xx, const __half2& yy)
                 {
-                    return __half2{x} += y;
+                    return __half2{xx} += yy;
                 }
                 friend
                 inline
                 __device__
-                __half2 operator-(const __half2& x, const __half2& y)
+                __half2 operator-(const __half2& xx, const __half2& yy)
                 {
-                    return __half2{x} -= y;
+                    return __half2{xx} -= yy;
                 }
                 friend
                 inline
                 __device__
-                __half2 operator*(const __half2& x, const __half2& y)
+                __half2 operator*(const __half2& xx, const __half2& yy)
                 {
-                    return __half2{x} *= y;
+                    return __half2{xx} *= yy;
                 }
                 friend
                 inline
                 __device__
-                __half2 operator/(const __half2& x, const __half2& y)
+                __half2 operator/(const __half2& xx, const __half2& yy)
                 {
-                    return __half2{x} /= y;
+                    return __half2{xx} /= yy;
                 }
                 friend
                 inline
                 __device__
-                bool operator==(const __half2& x, const __half2& y)
+                bool operator==(const __half2& xx, const __half2& yy)
                 {
-                    auto r = x.data == y.data;
+                    auto r = xx.data == yy.data;
                     return r.x != 0 && r.y != 0;
                 }
                 friend
                 inline
                 __device__
-                bool operator!=(const __half2& x, const __half2& y)
+                bool operator!=(const __half2& xx, const __half2& yy)
                 {
-                    return !(x == y);
+                    return !(xx == yy);
                 }
                 friend
                 inline
                 __device__
-                bool operator<(const __half2& x, const __half2& y)
+                bool operator<(const __half2& xx, const __half2& yy)
                 {
-                    auto r = x.data < y.data;
+                    auto r = xx.data < yy.data;
                     return r.x != 0 && r.y != 0;
                 }
                 friend
                 inline
                 __device__
-                bool operator>(const __half2& x, const __half2& y)
+                bool operator>(const __half2& xx, const __half2& yy)
                 {
-                    return y < x;
+                    return yy < xx;
                 }
                 friend
                 inline
                 __device__
-                bool operator<=(const __half2& x, const __half2& y)
+                bool operator<=(const __half2& xx, const __half2& yy)
                 {
-                    return !(y < x);
+                    return !(yy < xx);
                 }
                 friend
                 inline
                 __device__
-                bool operator>=(const __half2& x, const __half2& y)
+                bool operator>=(const __half2& xx, const __half2& yy)
                 {
-                    return !(x < y);
+                    return !(xx < yy);
                 }
             #endif // !defined(__HIP_NO_HALF_OPERATORS__)
         };
@@ -589,7 +581,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __high2half2(__half2 x)
             {
-                return __half2_raw{
+                return __half2{
                     _Float16_2{
                         static_cast<__half2_raw>(x).data.y,
                         static_cast<__half2_raw>(x).data.y}};
@@ -599,7 +591,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __lows2half2(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     _Float16_2{
                         static_cast<__half2_raw>(x).data.x,
                         static_cast<__half2_raw>(y).data.x}};
@@ -609,7 +601,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __highs2half2(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     _Float16_2{
                         static_cast<__half2_raw>(x).data.y,
                         static_cast<__half2_raw>(y).data.y}};
@@ -619,7 +611,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __lowhigh2highlow(__half2 x)
             {
-                return __half2_raw{
+                return __half2{
                     _Float16_2{
                         static_cast<__half2_raw>(x).data.y,
                         static_cast<__half2_raw>(x).data.x}};
@@ -712,7 +704,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __float2half2_rn(float x)
             {
-                return __half2_raw{
+                return __half2{
                     _Float16_2{
                         static_cast<_Float16>(x), static_cast<_Float16>(x)}};
             }
@@ -720,7 +712,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __floats2half2_rn(float x, float y)
             {
-                return __half2_raw{_Float16_2{
+                return __half2{_Float16_2{
                     static_cast<_Float16>(x), static_cast<_Float16>(y)}};
             }
             inline
@@ -1140,24 +1132,34 @@ THE SOFTWARE.
                 return static_cast<__half_raw>(x).data >
                     static_cast<__half_raw>(y).data;
             }
+            inline __device__
+            bool __hequ(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data < static_cast<__half_raw>(y).data) &&
+                    !(static_cast<__half_raw>(x).data > static_cast<__half_raw>(y).data);
+            }
+            inline __device__
+            bool __hneu(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data == static_cast<__half_raw>(y).data);
+            }
+            inline __device__
+            bool __hleu(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data > static_cast<__half_raw>(y).data);
+            }
             inline
             __device__
-            bool __hequ(__half x, __half y) { return __heq(x, y); }
+            bool __hgeu(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data < static_cast<__half_raw>(y).data);
+            }
             inline
             __device__
-            bool __hneu(__half x, __half y) { return __hne(x, y); }
+            bool __hltu(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data >= static_cast<__half_raw>(y).data);
+            }
             inline
             __device__
-            bool __hleu(__half x, __half y) { return __hle(x, y); }
-            inline
-            __device__
-            bool __hgeu(__half x, __half y) { return __hge(x, y); }
-            inline
-            __device__
-            bool __hltu(__half x, __half y) { return __hlt(x, y); }
-            inline
-            __device__
-            bool __hgtu(__half x, __half y) { return __hgt(x, y); }
+            bool __hgtu(__half x, __half y) {
+                return !(static_cast<__half_raw>(x).data <= static_cast<__half_raw>(y).data);
+            }
 
             inline
             __HOST_DEVICE__
@@ -1207,24 +1209,42 @@ THE SOFTWARE.
                     static_cast<__half2_raw>(y).data;
                 return __builtin_convertvector(-r, _Float16_2);
             }
+            inline __HOST_DEVICE__
+            __half2 __hequ2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data < static_cast<__half2_raw>(y).data) &&
+                    !(static_cast<__half2_raw>(x).data > static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
             inline
             __HOST_DEVICE__
-            __half2 __hequ2(__half2 x, __half2 y) { return __heq2(x, y); }
+            __half2 __hneu2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data == static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
             inline
             __HOST_DEVICE__
-            __half2 __hneu2(__half2 x, __half2 y) { return __hne2(x, y); }
+            __half2 __hleu2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data > static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
             inline
             __HOST_DEVICE__
-            __half2 __hleu2(__half2 x, __half2 y) { return __hle2(x, y); }
+            __half2 __hgeu2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data < static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
             inline
             __HOST_DEVICE__
-            __half2 __hgeu2(__half2 x, __half2 y) { return __hge2(x, y); }
+            __half2 __hltu2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data >= static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
             inline
             __HOST_DEVICE__
-            __half2 __hltu2(__half2 x, __half2 y) { return __hlt2(x, y); }
-            inline
-            __HOST_DEVICE__
-            __half2 __hgtu2(__half2 x, __half2 y) { return __hgt2(x, y); }
+            __half2 __hgtu2(__half2 x, __half2 y) {
+                auto r = !(static_cast<__half2_raw>(x).data <= static_cast<__half2_raw>(y).data);
+                return __builtin_convertvector(-r, _Float16_2);
+            }
 
             inline
             __HOST_DEVICE__
@@ -1408,7 +1428,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __hadd2(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     static_cast<__half2_raw>(x).data +
                     static_cast<__half2_raw>(y).data};
             }
@@ -1416,14 +1436,14 @@ THE SOFTWARE.
 	    __HOST_DEVICE__
 	    __half2 __habs2(__half2 x)
 	    {
-	        return __half2_raw{
+	        return __half2{
 		    __ocml_fabs_2f16(static_cast<__half2_raw>(x).data)};
 	    }
             inline
             __HOST_DEVICE__
             __half2 __hsub2(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     static_cast<__half2_raw>(x).data -
                     static_cast<__half2_raw>(y).data};
             }
@@ -1431,7 +1451,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __hmul2(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     static_cast<__half2_raw>(x).data *
                     static_cast<__half2_raw>(y).data};
             }
@@ -1466,7 +1486,7 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __hfma2(__half2 x, __half2 y, __half2 z)
             {
-                return __half2_raw{__ocml_fma_2f16(x, y, z)};
+                return __half2{__ocml_fma_2f16(x, y, z)};
             }
             inline
             __HOST_DEVICE__
@@ -1481,11 +1501,44 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 __h2div(__half2 x, __half2 y)
             {
-                return __half2_raw{
+                return __half2{
                     static_cast<__half2_raw>(x).data /
                     static_cast<__half2_raw>(y).data};
             }
 
+            // Atomic
+            #if defined(__clang__) && defined(__HIP__)
+            inline __device__ __half2 unsafeAtomicAdd(__half2* address, __half2 value) {
+            #if __has_builtin(__builtin_amdgcn_flat_atomic_fadd_v2f16)
+                // The api expects an ext_vector_type of half
+                typedef _Float16 __attribute__((ext_vector_type(2))) vec_fp162;
+                static_assert(sizeof(vec_fp162) == sizeof(__half2_raw));
+                union {
+                    __half2_raw h2r;
+                    vec_fp162 fp16;
+                } u {static_cast<__half2_raw>(value)};
+                u.fp16 =
+                    __builtin_amdgcn_flat_atomic_fadd_v2f16((vec_fp162*)address, u.fp16);
+                return static_cast<__half2>(u.h2r);
+            #else
+                static_assert(sizeof(__half2_raw) == sizeof(unsigned int));
+                union u_hold {
+                    __half2_raw h2r;
+                    unsigned int u32;
+                };
+                u_hold old_val, new_val;
+                old_val.u32 = __hip_atomic_load((unsigned int*)address, __ATOMIC_RELAXED,
+                                                __HIP_MEMORY_SCOPE_AGENT);
+                do {
+                    new_val.h2r = __hadd2(old_val.h2r, value);
+                } while (!__hip_atomic_compare_exchange_strong(
+                    (unsigned int*)address, &old_val.u32, new_val.u32, __ATOMIC_RELAXED,
+                    __ATOMIC_RELAXED, __HIP_MEMORY_SCOPE_AGENT));
+                return old_val.h2r;
+            #endif
+            }
+            #endif // defined(__clang__) && defined(__HIP__)
+
             // Math functions
             #if defined(__clang__) && defined(__HIP__)
             inline
@@ -1585,7 +1638,7 @@ THE SOFTWARE.
             __half hrcp(__half x)
             {
                 return __half_raw{
-                    static_cast<_Float16>(__builtin_amdgcn_rcph(static_cast<__half_raw>(x).data))};
+                    static_cast<_Float16>(1.0f) /static_cast<__half_raw>(x).data};
             }
             inline
             __device__
@@ -1624,61 +1677,61 @@ THE SOFTWARE.
             __HOST_DEVICE__
             __half2 h2trunc(__half2 x)
             {
-                return __half2_raw{__ocml_trunc_2f16(x)};
+                return __half2{__ocml_trunc_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2ceil(__half2 x)
             {
-                return __half2_raw{__ocml_ceil_2f16(x)};
+                return __half2{__ocml_ceil_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2floor(__half2 x)
             {
-                return __half2_raw{__ocml_floor_2f16(x)};
+                return __half2{__ocml_floor_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2rint(__half2 x)
             {
-                return __half2_raw{__ocml_rint_2f16(x)};
+                return __half2{__ocml_rint_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2sin(__half2 x)
             {
-                return __half2_raw{__ocml_sin_2f16(x)};
+                return __half2{__ocml_sin_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2cos(__half2 x)
             {
-                return __half2_raw{__ocml_cos_2f16(x)};
+                return __half2{__ocml_cos_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2exp(__half2 x)
             {
-                return __half2_raw{__ocml_exp_2f16(x)};
+                return __half2{__ocml_exp_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2exp2(__half2 x)
             {
-                return __half2_raw{__ocml_exp2_2f16(x)};
+                return __half2{__ocml_exp2_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2exp10(__half2 x)
             {
-                return __half2_raw{__ocml_exp10_2f16(x)};
+                return __half2{__ocml_exp10_2f16(x)};
             }
             inline
             __HOST_DEVICE__
             __half2 h2log2(__half2 x)
             {
-                return __half2_raw{__ocml_log2_2f16(x)};
+                return __half2{__ocml_log2_2f16(x)};
             }
             inline
             __HOST_DEVICE__
@@ -1689,8 +1742,8 @@ THE SOFTWARE.
             inline
             __HOST_DEVICE__
             __half2 h2rcp(__half2 x) {
-                return _Float16_2{static_cast<_Float16>(__builtin_amdgcn_rcph(x.x)),
-                                  static_cast<_Float16>(__builtin_amdgcn_rcph(x.y))};
+                return _Float16_2{
+                    _Float16_2{static_cast<_Float16>(1.0f), static_cast<_Float16>(1.0f)} / x.data};
             }
             inline
             __HOST_DEVICE__
@@ -1703,7 +1756,7 @@ THE SOFTWARE.
             __half2 __hisinf2(__half2 x)
             {
                 auto r = __ocml_isinf_2f16(x);
-                return __half2_raw{_Float16_2{
+                return __half2{_Float16_2{
                     static_cast<_Float16>(r.x), static_cast<_Float16>(r.y)}};
             }
             inline
@@ -1711,14 +1764,14 @@ THE SOFTWARE.
             __half2 __hisnan2(__half2 x)
             {
                 auto r = __ocml_isnan_2f16(x);
-                return __half2_raw{_Float16_2{
+                return __half2{_Float16_2{
                     static_cast<_Float16>(r.x), static_cast<_Float16>(r.y)}};
             }
             inline
             __HOST_DEVICE__
             __half2 __hneg2(__half2 x)
             {
-                return __half2_raw{-static_cast<__half2_raw>(x).data};
+                return __half2{-static_cast<__half2_raw>(x).data};
             }
         } // Anonymous namespace.
 
@@ -1783,14 +1836,10 @@ THE SOFTWARE.
              return tmp.h;
          }
     #endif // defined(__cplusplus)
-#elif defined(__GNUC__)
+#elif defined(__GNUC__) || defined(_MSC_VER)
     #if !defined(__HIPCC_RTC__)
       #include "hip_fp16_gcc.h"
     #endif
 #endif // !defined(__clang__) && defined(__GNUC__)
 
-#if defined(__clang__)
-#pragma clang diagnostic pop
-#endif
-
 #endif // HIP_INCLUDE_HIP_AMD_DETAIL_HIP_FP16_H
